#!/usr/bin/perl
# Perl script to compute BLEU score given a file containing segment-level
# information. This is not required for significance testing, but can be used to  
# check whether segment-level scores were correctly output. The scores computed using 
# this script should be identical to the scores output by the mteval-v13a.pl script.
#
# Kevin Gimpel
# 4/2/2011

if (scalar(@ARGV) != 1) {
    print "Usage: compute_bleu_from_stats_v13a.pl bleu_stats_file\n";
    exit;
}

my $max_Ngram = 9;
my ($cum_ref_length, @cum_match, @cum_tst_cnt, @cum_ref_cnt, @cum_tst_info, @cum_ref_info);
my $BLEU_BP = \&brevity_penalty_closest;
 
$cum_ref_length = 0;
for (my $j=1; $j<=$max_Ngram; $j++) {
    $cum_match[$j] = $cum_tst_cnt[$j] = $cum_ref_cnt[$j] = $cum_tst_info[$j] = $cum_ref_info[$j] = 0;
}
open(IN1, "<", $ARGV[0]);
my $sc_ref_sys;
while (<IN1>) {
	$line = $_;
	chomp($line);

	my ($shortest_ref_length, @match_cnt, @tst_cnt, @ref_cnt, @tst_info, @ref_info);
	$shortest_ref_length = 0;
	for (my $j=1; $j<=$max_Ngram; $j++) {
		$match_cnt[$j] = $tst_cnt[$j] = $ref_cnt[$j] = $tst_info[$j] = $ref_info[$j] = 0;
	}


	@pieces = split /\s*\|\s*/, $line; 
	$sc_ref_sys = $pieces[0];
	$shortest_ref_length = $pieces[1];
	
	$match_cnt_line = $pieces[2];
	@subpieces = split / /, $match_cnt_line;
	for (my $j=1; $j<= $max_Ngram; $j++) {
		$match_cnt[$j] = @subpieces[$j - 1];
	}
	$tst_cnt_line = $pieces[3];
	@subpieces = split / /, $tst_cnt_line;
	for (my $j=1; $j<= $max_Ngram; $j++) {
		$tst_cnt[$j] = @subpieces[$j - 1];
	}

	$ref_cnt_line = $pieces[4];
	@subpieces = split / /, $ref_cnt_line;
	for (my $j=1; $j<= $max_Ngram; $j++) {
		$ref_cnt[$j] = @subpieces[$j - 1];
	}
	$tst_info_line = $pieces[5];
	@subpieces = split / /, $tst_info_line;
	for (my $j=1; $j<= $max_Ngram; $j++) {
		$tst_info[$j] = @subpieces[$j - 1];
	}
	$ref_info_line = $pieces[6];
	@subpieces = split / /, $ref_info_line;
	for (my $j=1; $j<= $max_Ngram; $j++) {
		$ref_info[$j] = @subpieces[$j - 1];
	}
	$sys = $pieces[7];
        $cum_ref_length += $shortest_ref_length;
        for (my $j=1; $j<=$max_Ngram; $j++) {
            $cum_match[$j] += $match_cnt[$j];
            $cum_tst_cnt[$j] += $tst_cnt[$j];
            $cum_ref_cnt[$j] += $ref_cnt[$j];
            $cum_tst_info[$j] += $tst_info[$j];
            $cum_ref_info[$j] += $ref_info[$j];
        }
	
}

close IN1;

# my %DOCmt = ();
# printf "NIST score using   5-grams = %.4f\n",
# nist_score ($sc_ref_sys, \@cum_match, \@cum_tst_cnt, \@cum_ref_cnt, \@cum_tst_info, \@cum_ref_info, $sys, %DOCmt);
my %DOCmt = ();
printf "BLEU score using   4-grams = %.4f\n",
bleu_score($cum_ref_length, \@cum_match, \@cum_tst_cnt, $sys, %DOCmt);
 
#($date, $time) = date_time_stamp();
#print "MT evaluation scorer ended on $date at $time\n";
 
exit 0;




sub score_segment {
 
    my ($tst_seg, @ref_segs) = @_;
    my (@tst_wrds, %tst_ngrams, @match_count, @tst_count, @tst_info);
    my (@ref_wrds, $ref_seg, %ref_ngrams, %ref_ngrams_max, @ref_count, @ref_info);
    my ($ngram);
    my (@nwrds_ref);
    my $ref_length;
 
    for (my $j=1; $j<= $max_Ngram; $j++) {
        $match_count[$j] = $tst_count[$j] = $ref_count[$j] = $tst_info[$j] = $ref_info[$j] = 0;
    }
 
# get the ngram counts for the test segment
    @tst_wrds = split /\s+/, $tst_seg;
    %tst_ngrams = %{Words2Ngrams (@tst_wrds)};
    for (my $j=1; $j<=$max_Ngram; $j++) { # compute ngram counts
        $tst_count[$j]  = $j<=@tst_wrds ? (@tst_wrds - $j + 1) : 0;
    }
 
# get the ngram counts for the reference segments
    foreach $ref_seg (@ref_segs) {
        @ref_wrds = split /\s+/, $ref_seg;
        %ref_ngrams = %{Words2Ngrams (@ref_wrds)};
        foreach $ngram (keys %ref_ngrams) { # find the maximum # of occurrences
            my @wrds = split / /, $ngram;
            $ref_info[@wrds] += $ngram_info{$ngram};
            $ref_ngrams_max{$ngram} = defined $ref_ngrams_max{$ngram} ?
                max ($ref_ngrams_max{$ngram}, $ref_ngrams{$ngram}) :
                    $ref_ngrams{$ngram};
        }
        for (my $j=1; $j<=$max_Ngram; $j++) { # update ngram counts
            $ref_count[$j] += $j<=@ref_wrds ? (@ref_wrds - $j + 1) : 0;
        }
	if ( not defined( $ref_length ) )
	{
		$ref_length = scalar( @ref_wrds );
	}
	else
	{
		$ref_length = &{$BLEU_BP}( $ref_length, scalar( @ref_wrds ), scalar( @tst_wrds ) );
	}
    }
 
# accumulate scoring stats for tst_seg ngrams that match ref_seg ngrams
    foreach $ngram (keys %tst_ngrams) {
        next unless defined $ref_ngrams_max{$ngram};
        my @wrds = split / /, $ngram;
        $tst_info[@wrds] += $ngram_info{$ngram} * min($tst_ngrams{$ngram},$ref_ngrams_max{$ngram});
        $match_count[@wrds] += my $count = min($tst_ngrams{$ngram},$ref_ngrams_max{$ngram});
        printf "%.2f info for each of $count %d-grams = '%s'\n", $ngram_info{$ngram}, scalar @wrds, $ngram
            if $detail >= 3;
    }
 
    return ($ref_length, [@match_count], [@tst_count], [@ref_count], [@tst_info], [@ref_info]);
}


###############################################################################################################################
# Default method used to compute the BLEU score, using smoothing.
# Note that the method used can be overridden using the '--no-smoothing' command-line argument
# The smoothing is computed by taking 1 / ( 2^k ), instead of 0, for each precision score whose matching n-gram count is null
# k is 1 for the first 'n' value for which the n-gram match count is null
# For example, if the text contains:
#   - one 2-gram match
#   - and (consequently) two 1-gram matches
# the n-gram count for each individual precision score would be:
#   - n=1  =>  prec_count = 2     (two unigrams)
#   - n=2  =>  prec_count = 1     (one bigram)
#   - n=3  =>  prec_count = 1/2   (no trigram,  taking 'smoothed' value of 1 / ( 2^k ), with k=1)
#   - n=4  =>  prec_count = 1/4   (no fourgram, taking 'smoothed' value of 1 / ( 2^k ), with k=2)
###############################################################################################################################
sub bleu_score
{
	my ($ref_length, $matching_ngrams, $tst_ngrams, $sys, $SCOREmt) = @_;
	my $score = 0;
	my $iscore = 0;
	my $exp_len_score = 0;
	$exp_len_score = exp( min (0, 1 - $ref_length / $tst_ngrams->[ 1 ] ) ) if ( $tst_ngrams->[ 1 ] > 0 );
	my $smooth = 1;
	for ( my $j = 1; $j <= $max_Ngram; ++$j )
	{
		if ( $tst_ngrams->[ $j ] == 0 )
		{
			$iscore = 0;
		}
		elsif ( $matching_ngrams->[ $j ] == 0 )
		{
			$smooth *= 2;
			$iscore = log( 1 / ( $smooth * $tst_ngrams->[ $j ] ) );
		}
		else
		{
			$iscore = log( $matching_ngrams->[ $j ] / $tst_ngrams->[ $j ] );
		}
		$SCOREmt->{ $j }{ $sys }{ ind } = exp( $iscore );
		$score += $iscore;
		$SCOREmt->{ $j }{ $sys }{ cum } = exp( $score / $j ) * $exp_len_score;
	}
	return $SCOREmt->{ 4 }{ $sys }{ cum };
}

#################################

sub nist_score
{
	my ($nsys, $matching_ngrams, $tst_ngrams, $ref_ngrams, $tst_info, $ref_info, $sys, $SCOREmt) = @_;
	my $score = 0;
	my $iscore = 0;

	for (my $n=1; $n<=$max_Ngram; $n++)
	{
		$score += $tst_info->[$n]/max($tst_ngrams->[$n],1);
		$SCOREmt->{$n}{$sys}{cum} = $score * nist_length_penalty($tst_ngrams->[1]/($ref_ngrams->[1]/$nsys));
		$iscore = $tst_info->[$n]/max($tst_ngrams->[$n],1);
		$SCOREmt->{$n}{$sys}{ind} = $iscore * nist_length_penalty($tst_ngrams->[1]/($ref_ngrams->[1]/$nsys));
	}
	return $SCOREmt->{5}{$sys}{cum};
}

#################################

sub nist_length_penalty
{
	my ($ratio) = @_;
	return 1 if $ratio >= 1;
	return 0 if $ratio <= 0;
	my $ratio_x = 1.5;
	my $score_x = 0.5;
	my $beta = -log($score_x)/log($ratio_x)/log($ratio_x);
	return exp (-$beta*log($ratio)*log($ratio));
}

#################################
 
sub date_time_stamp {
 
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime();
    my @months = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
    my ($date, $time);
                                                                                                                                                    
    $time = sprintf "%2.2d:%2.2d:%2.2d", $hour, $min, $sec;
    $date = sprintf "%4.4s %3.3s %s", 1900+$year, $months[$mon], $mday;
    return ($date, $time);
}
 
#################################
 
sub max {
 
    my ($max, $next);
 
    return unless defined ($max=pop);
    while (defined ($next=pop)) {
        $max = $next if $next > $max;
    }
    return $max;
}
 
#################################
 
sub min {
 
    my ($min, $next);
 
    return unless defined ($min=pop);
    while (defined ($next=pop)) {
        $min = $next if $next < $min;
    }
    return $min;
}
 
#################################
 
sub printout_report
{
  
    if ( $METHOD eq "BOTH" ) {
        foreach my $sys (sort @tst_sys) {
            printf "NIST score = %2.4f  BLEU score = %.4f for system \"$sys\"\n",$NISTmt{5}{$sys}{cum},$BLEUmt{4}{$sys}{cum};
        }
    } elsif ($METHOD eq "NIST" ) {
        foreach my $sys (sort @tst_sys) {
            printf "NIST score = %2.4f  for system \"$sys\"\n",$NISTmt{5}{$sys}{cum};
        }
    } elsif ($METHOD eq "BLEU" ) {
        foreach my $sys (sort @tst_sys) {
            printf "\nBLEU score = %.4f for system \"$sys\"\n",$BLEUmt{4}{$sys}{cum};
        }
    }
     
  
    printf "\n# ------------------------------------------------------------------------\n\n";
    printf "Individual N-gram scoring\n";
    printf "        1-gram   2-gram   3-gram   4-gram   5-gram   6-gram   7-gram   8-gram   9-gram\n";
    printf "        ------   ------   ------   ------   ------   ------   ------   ------   ------\n";
                                                                                                                                                    
    if (( $METHOD eq "BOTH" ) || ($METHOD eq "NIST")) {
        foreach my $sys (sort @tst_sys) {
            printf " NIST:";
            for (my $i=1; $i<=$max_Ngram; $i++) {
                printf "  %2.4f ",$NISTmt{$i}{$sys}{ind}
            }
            printf " \"$sys\"\n";
        }
        printf "\n";
    }
     
    if (( $METHOD eq "BOTH" ) || ($METHOD eq "BLEU")) {
        foreach my $sys (sort @tst_sys) {
            printf " BLEU:";
            for (my $i=1; $i<=$max_Ngram; $i++) {
                printf "  %2.4f ",$BLEUmt{$i}{$sys}{ind}
            }
           printf " \"$sys\"\n";
        }
    }
     
    printf "\n# ------------------------------------------------------------------------\n";
    printf "Cumulative N-gram scoring\n";
    printf "        1-gram   2-gram   3-gram   4-gram   5-gram   6-gram   7-gram   8-gram   9-gram\n";
    printf "        ------   ------   ------   ------   ------   ------   ------   ------   ------\n";
  
    if (( $METHOD eq "BOTH" ) || ($METHOD eq "NIST")) {
        foreach my $sys (sort @tst_sys) {
            printf " NIST:";
            for (my $i=1; $i<=$max_Ngram; $i++) {
                printf "  %2.4f ",$NISTmt{$i}{$sys}{cum}
            }
            printf " \"$sys\"\n";
        }
    }
    printf "\n";
                                                                                                                                                    
  
    if (( $METHOD eq "BOTH" ) || ($METHOD eq "BLEU")) {
        foreach my $sys (sort @tst_sys) {
            printf " BLEU:";
            for (my $i=1; $i<=$max_Ngram; $i++) {
                printf "  %2.4f ",$BLEUmt{$i}{$sys}{cum}
            }
            printf " \"$sys\"\n";
        }
    }
}

###############################################################################################################################
# function returning the closest reference length (to the candidate sentence length)
# takes as input:
#  - currentLength: the current (closest) reference length.
#  - candidateSentenceLength : the current reference sentence length
#  - candidateSentenceLength : the current candidate sentence length
# when two reference sentences are at the same distance, it will return the shortest reference sentence length
# example of 4 iterations, given:
#  - one candidate sentence containing 7 tokens
#  - one reference translation containing 11 tokens
#  - one reference translation containing 8 tokens
#  - one reference translation containing 6 tokens
#  - one reference translation containing 7 tokens
# the multiple invokations will return:
#  - currentLength is set to 11 (outside of this function)
#  - brevity_penalty_closest( 11, 8, 7 ) returns 8, since abs( 8 - 7 ) < abs( 11 - 7 )
#  - brevity_penalty_closest( 8, 6, 7 ) returns 6, since abs( 8 - 7 ) == abs( 6 - 7 ) AND 6 < 8
#  - brevity_penalty_closest( 7, 6, 7 ) returns 7, since abs( 7 - 7 ) < abs( 6 - 7 )
###############################################################################################################################
sub brevity_penalty_closest
{
	my ( $currentLength, $referenceSentenceLength, $candidateSentenceLength ) = @_;
	my $result = $currentLength;
	if ( abs( $candidateSentenceLength - $referenceSentenceLength ) <= abs( $candidateSentenceLength - $currentLength ) )
	{
		if ( abs( $candidateSentenceLength - $referenceSentenceLength ) == abs( $candidateSentenceLength - $currentLength ) )
		{
			if ( $currentLength > $referenceSentenceLength )
			{
				$result = $referenceSentenceLength;
			}
		}
		else
		{
			$result = $referenceSentenceLength;
		}
	}
	return $result;
}
